#!/bin/bash
#
# This is the script I use to build a full Docker-less NickelTC sysroot.
# What's in here is perfectly sane and tested,
# what might fail is the kobo-nickel-sysroot script, probably because of stupid typos,
# since I've kept it updated, but have never really rebuilt it since the last NickelTC update.
#
# Speaking of, the Docker variant of this, maintained by @pgaskin, is available over here:
# <https://github.com/pgaskin/NickelTC>.
#
# $Id: nickeltc_build.sh 18708 2021-08-29 14:52:32Z NiLuJe $
#
##

set -euo pipefail

if [[ $# -ne 1 ]]; then
    echo "Usage: $0 OUTPUT_PATH"
    echo "I strongly recommend that that output directory be in a (large) tmpfs!"
    exit 2
fi

# NOTE: Make that a warning, to allow us to point to an existing tmpfs
if [[ -d "$1" ]]; then
    echo "Warning: Output '$1' already exists"
fi

echo "> Setting up output directory '$1'"
tc_path="$(readlink -f "$1")"
mkdir -pv "${tc_path}"
cd "${tc_path}"

# Start by building the TC via koxtoolchain
echo "> Building the TC"
(
	# NOTE: This compiles everything in $PWD
	rm -rf koxtoolchain
	until git clone --depth 1 https://github.com/koreader/koxtoolchain.git ; do
		rm -rf koxtoolchain
		sleep 15
	done
	cd koxtoolchain
	./gen-tc.sh nickel
)

echo "> Cleaning up the TC build tree"
rm -rf koxtoolchain

# Then Qt
echo "> Building Qt"
(
	# NOTE: This will build in ${HOME}/Kobo/CrossTool
	#       I recommend setting that up as a tmpfs... 10G works. Less probably will, too.
	svn cat https://svn.ak-team.com/svn/Configs/trunk/Kindle/Misc/kobo-nickel-sysroot.sh > kobo-nickel-sysroot.sh
	chmod -cvR a+x kobo-nickel-sysroot.sh
	# We'll also need the CMake Cross file...
	svn cat https://svn.ak-team.com/svn/Configs/trunk/Kindle/Misc/CMakeCross.txt > CMakeCross.txt
	./kobo-nickel-sysroot.sh
)

echo "> Merging the Qt sysroot with the TC"
find "${HOME}/x-tools/arm-nickel-linux-gnueabihf" -type d -exec chmod a+w '{}' \+
cp -av "${HOME}/Kobo/CrossTool/Build_NICKEL/arm-nickel-linux-gnueabihf/." "${HOME}/x-tools/arm-nickel-linux-gnueabihf/."
find "${HOME}/x-tools/arm-nickel-linux-gnueabihf" -type d -exec chmod a-w '{}' \+

echo "> Cleaning up the Qt build tree"
rm -rf "${HOME}/Kobo/CrossTool/Build_NICKEL"

echo "> Relocating Qt"
chmod a+w "${HOME}/x-tools/arm-nickel-linux-gnueabihf/bin"
cat <<-\EOF | tee "${HOME}/x-tools/arm-nickel-linux-gnueabihf/bin/qt.conf"
[Paths]
Prefix = ../arm-nickel-linux-gnueabihf/sysroot/usr
EOF
chmod a-w "${HOME}/x-tools/arm-nickel-linux-gnueabihf/bin"

# Now, remove build-time paths from the mkspecs, too
echo "> Cleaning up qmake spec"
chmod a+w ${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr/mkspecs/linux-arm-nickel-gnueabihf-g++
sed '0,/^QMAKE_CFLAGS/{//d;}' -i ${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr/mkspecs/linux-arm-nickel-gnueabihf-g++/qmake.conf
sed '0,/^QMAKE_CXXFLAGS/{//d;}' -i ${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr/mkspecs/linux-arm-nickel-gnueabihf-g++/qmake.conf
sed '0,/^QMAKE_LFLAGS/{//d;}' -i ${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr/mkspecs/linux-arm-nickel-gnueabihf-g++/qmake.conf
sed -e "s:${HOME}/Kobo/CrossTool/Build_NICKEL/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr:${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr:g" -i ${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr/mkspecs/linux-arm-nickel-gnueabihf-g++/qmake.conf
chmod a-w ${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr/mkspecs/linux-arm-nickel-gnueabihf-g++

# libtool is the worst. .la files are actively harmful on Linux, c.f., https://flameeyes.blog/2008/04/14/what-about-those-la-files/
echo "> Neutering libtool madness"
chmod a+w ${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr/lib
find "${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr/lib" -name "*.la" -type f -delete
# Handle's QMake's own insane dependency management thingy, too, while we're there.
sed -e "s:${HOME}/Kobo/CrossTool/Build_NICKEL/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr:${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr:g" -i ${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr/lib/*.prl
chmod a-w ${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr/lib

# Handle pkg-config shenanigans, c.f., https://github.com/pgaskin/kobo-plugin-experiments/commit/7020977c611c9301c07ef1cb24656fd09acef77a
echo "> Relocating pkg-config"
# First, get rid of the build time prefix (as well as any stray build time path that made it to a hard-coded -I or -L entry...)
chmod -R u+w "${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr/lib/pkgconfig"
sed -e "s:${HOME}/Kobo/CrossTool/Build_NICKEL/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr:${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr:g" -i ${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr/lib/pkgconfig/*.pc
chmod -R u-w "${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot/usr/lib/pkgconfig"
# Then, create a triplet-prefixed pkg-config wrapper (c.f., https://autotools.io/pkgconfig/cross-compiling.html).
# autotools should pick it up automagically. Otherwise, enforce it via the PKG_CONFIG env var.
chmod a+w "${HOME}/x-tools/arm-nickel-linux-gnueabihf/bin"
cat <<-\EOF | tee "${HOME}/x-tools/arm-nickel-linux-gnueabihf/bin/arm-nickel-linux-gnueabihf-pkg-config"
#!/bin/sh

SYSROOT="${HOME}/x-tools/arm-nickel-linux-gnueabihf/arm-nickel-linux-gnueabihf/sysroot"

export PKG_CONFIG_PATH=
export PKG_CONFIG_LIBDIR="${SYSROOT}/usr/lib/pkgconfig:${SYSROOT}/usr/share/pkgconfig"
# NOTE: I prefer relying on --define-prefix rather than enforcing a fixed sysroot, for a simple reason:
#       that sysroot is ro, so you should be able to build and install stuff in a secondary staging directory,
#       and have pkg-config handle that automagically simply by expanding the PKG_CONFIG_LIBDIR search path.
#       You can either tweak this wrapper, or bypass it and setup the various pkg-config env vars yourself,
#       c.f., http://trac.ak-team.com/trac/browser/niluje/Configs/trunk/Kindle/Misc/x-compile.sh
# NOTE: More importantly, we've already modified our .pc files so that prefix already points to ${SYSROOT}!
#       Which is why we enforce an empty one here, as we generally expect the sysroot's internal prefix to be /usr.
#       (i.e., another less flexible approach would be to post-process the .pc files to set prefix to /usr, and point
#       PKG_CONFIG_SYSROOT_DIR to $SYSROOT).
export PKG_CONFIG_SYSROOT_DIR=""
# Let pkg-config strip the right redundant system paths
export PKG_CONFIG_SYSTEM_INCLUDE_PATH="${SYSROOT}/usr/include"
export PKG_CONFIG_SYSTEM_LIBRARY_PATH="${SYSROOT}/usr/lib:${SYSROOT}/lib"

exec pkg-config --define-prefix "$@"
EOF
chmod a-w "${HOME}/x-tools/arm-nickel-linux-gnueabihf/bin"
chmod a+rx "${HOME}/x-tools/arm-nickel-linux-gnueabihf/bin/arm-nickel-linux-gnueabihf-pkg-config"

echo "> Done."
echo "> You can now prepend ${HOME}/x-tools/arm-nickel-linux-gnueabihf/bin to your PATH:"
echo "> export PATH='${HOME}/x-tools/arm-nickel-linux-gnueabihf/bin:\${PATH}'"
echo "> Alternatively, you can use the x-compile.sh script to setup a full cross-compilation env:"
echo "> source ${PWD}/x-compile.sh nickel env"
